/* Morphological Analysis of Korean
 *
 * create : 95.5.31 (Wed), bgjang@csone
 * modify :
 * 95.6.1 (Thu), bgjang@csone - main skeleton
 * 95.6.2 (Fri), bgjang@csone - searching step of normal dict
 * 95.6.4 (Sun), bgjang@csone - complete searching step of normal
 *                             dict, Result edge pool (chart),
 *                             result print routine
 * 95.6.18 (Sun), bgjang@csone - doIrrExecWordSearch()
 * 95.6.19 (Mon), bgjang@csone - makePTypeCharts()
 * 95.6.24-26 (Mon), bgjang@csone - normal word searching step, and results
 *                                 printing routines. And clear the source
 *                                 code cause MAchart*() are rewritten.
 * 95.6.26 (Mon), bgjang@csone - version 0.5 alpha
 * 95.7.18 (Tue), bgjang - All sub jobs are moved to MAdo.c file
 * 95.7.19-31 (Mon), bgjang 
 *    environment variables are added. mTable for unknown word processing
 *    is added. bugs using MAchartDestroy() is cleaned.
 *
 *
 *------------------------------------------------------------------
 *
 * preprocessing :
 * 1. tokenization and make list of token for one sentence which
 *    is terminated by newline.
 * 2. decide word, currently we choose word is terminated by
 *    space characters, so ignore the error of wrongly used space.
 *
 * rough algorithm :
 * - check the potential part of string in which irregular and
 *  exception can be occured.
 * - make charts which have potential path that can be generated
 *  by the first step's part of string's changes.
 * - for each chart generated in second step, do normal word
 *  searching and print the result.
 *
 * Mainly unwritten part :
 * - Using syllable information : this program don't use syllable information
 *  to decrease the dictionary lookup frequency.
 * - Checking the ending of word when checking irregular part of string :
 *  irregulation only can be occured in the ending of word, so this
 *  informations can be used to decrease the generation of second step's
 *  charts.
 *
 *------------------------------------------------------------------
 */
#include <stdio.h>
#include <assert.h>
#include <fcntl.h>
#include <MAdict.h>
#include "version.h"
#include "common.h"
#include "pos.h"
#include "MAchart.h"
#include "MAerror.h"
#include <stdlib.h>
#include <limits.h>


/* */
/* default dictionary names to be used */
#ifndef DEFAULT_NDICTNAME
#define DEFAULT_NDICTNAME   "nn"
#endif
#ifndef DEFAULT_EDICTNAME
#define DEFAULT_EDICTNAME   "ee"
#endif
#ifndef DEFAULT_IDICTNAME
#define DEFAULT_IDICTNAME   "ii"
#endif
#ifndef DEFAULT_UDICTNAME
#define DEFAULT_UDICTNAME   "uu"
#endif


/* static function declarations -------------------------------------*/
static void prepareProcess();
static void checkDictOpen();


/* extern variables -------------------------------------------------*/
// what is initialized ?
int isInitDict = FALSE;
int isInitToken = FALSE;
int isInitChart = FALSE;
int isInitStack = FALSE;

int isFileOpened = FALSE;
int isUsingCache = TRUE;

// which dictionary ?
int NdictID = -1;       // normal dictionary
int EdictID = -1;       // execption dictionary
int UdictID = -1;       // unknown word dictionary
int IdictID = -1;       // irregular dictionary

/* if the input for MoA is from file, then this pointer
 * point to that file's file pointer.
 * It must be defined as global cause tokenization routines
 * use this variables as global.
 */
FILE* fpInput;

/* connectivity table - which POS can be connected to which POS ?
 */
char cTable[NUM_POS][NUM_POS];

/* connectivity table in Unknown word processing
 * it is more restricted than connectivity table.
 */
char mTable[NUM_POS][NUM_POS];

/* dictionary must be located in this directory */
static char dictDir[_POSIX_PATH_MAX];

/* mTable and cTable must be localted in this directory */
static char tableDir[_POSIX_PATH_MAX];


#if defined(LOG_STATISTIC)
int maxUsedChartNum = 0;
#endif


main( argc, argv )
    int argc;
    char* argv[];
{
    int r;
    int PchartIDStackID;
    int TailVStackID;
    int SchartID;
    int TchartID;
    int PchartID;
    int RchartID;
    int isInterpreted = FALSE;
    int lastTailV; /* the first head vertex ID is implicitly known to
                    * the other functions as Zero. And the last tail vertex ID
                    * is variable so must be recorded.
                    */
#if defined(DEBUG)
    int countPchart = 0;
#endif

    setFuncName( "main" );

    prepareProcess( argc, argv );

    /* get list of tokens for one sentence
     * and do something
     */
    while(1) {

        // result in (aToken*) tokenList, (int) tokenNumber
        if( !tokenization() ) break;

        debugMesg("\nCHECK : TOKENIZATION for one sentence.\n\n");

        while(1) {

            /* create S-Type chart to store syllable */
            MAchartCreate( &SchartID );

            /* select one word and fill in S-type chart identified by
             * SchartID and the last tail vertex ID will be returned.
             */
            if( !selectOneWordAndMakeChart( SchartID, &lastTailV ) ) break;

            // MAchartDump( SchartID, 1, 1, 1 );   // debugging

            debugMesg("\nCHECK : selectOneWordAndMakeChart() - %d lastTailV.\n\n", lastTailV);

            /* create T-Type chart to store the irr/excep position */
            MAchartCreate( &TchartID );

            /* result in TchartID - check Irregular and Exception in word
             * and make one T-type chart that is identified by TchartID.
             */
            doIrrExecWordSearch( SchartID, TchartID, lastTailV );

            debugMesg("\nCHECK : doIrrExecWordSearch() - end.\n" );
#if defined(DEBUG)
            MAchartDump( TchartID, 0, 0, 0 );
#endif

            /* it do DFS style searching for SchartID to make several
             * P-type chart which is processed directly to make
             * R-type chart.
             */
            makePTypeCharts( SchartID, TchartID, lastTailV, &PchartIDStackID );

            debugMesg("\nCHECK : makePTypeCharts() is done, with stack ID is %d.\n", PchartIDStackID );

            /* for the P-Type charts of which ID is stored in
             * stack, do normal word searching and print the results.
             *
             * error handling in each functions or cannot be
             * occured.
             */
            while( !isIsEmpty( PchartIDStackID ) ) {

                /* make R-Type chart which is used to store the
                 * result edges from normal searching step.
                 */
                MAchartCreate( &RchartID );

                isPop( PchartIDStackID, &PchartID );

#if defined(DEBUG)
                debugMesg("\nCHECK : normal word searching - [%d]th.\n", countPchart);
                countPchart++;
                debugMesg("\nThe followings are one of P-Type chart.\n" );
                MAchartDump( PchartID, 0, 0, 0 );
#endif

                r = doNormalWordSearch( PchartID, RchartID, lastTailV );

                if( r == TRUE ) {
                    char ch = 'o';
                    long irrApplied;
#if defined(DEBUG)
                    debugMesg("\nCHECK : searching was succeeded.\n");
                    MAchartDump( RchartID, 0, 0, 0 );
#endif
                    isInterpreted = TRUE;
                    MAchartGetIrrApplied( PchartID, &irrApplied );
                    doPrintResult( RchartID, lastTailV, irrApplied, ch );
                }
#if defined(DEBUG)
                else
                    debugMesg("\nCHECK : searching failed.\n");
#endif

                /* if unknown word processing step is needed, then
                 * the step will be located in here or above.
                 */

                /* destroy R-Type chart cause now R-Type chart must be
                 * initialized. And the last RchartID must be kept for
                 * Unknown word processing.
                 */
                if( !isIsEmpty(PchartIDStackID) || isInterpreted == TRUE ) {
                    MAchartDestroy( RchartID );
                    RchartID = -1;
                }
                /* if P-Type chart must be kept, then this coding must
                 * be changed
                 */
                if( isInterpreted == TRUE ) {
                    MAchartDestroy( PchartID );
                    PchartID = -1;
                }
            }
            isDestroy( PchartIDStackID );

#if defined(DEBUG)
            countPchart = 0;
#endif

            /* Unknown word processing for current PchartID, RchartID
             * This part is implcitly and strongly connected to
             * makePTypeCharts() part which makes the P-type charts.
             * In that functions, lastly the charts which includes no
             * irregular/exceptional is generated. And only this
             * chart is used for unknown word processing.
             */
            if( isInterpreted == FALSE ) {
                char ch = 'x';
                long irrApplied;

#if defined(DEBUG) || defined(DEBUGU)
                MAchartDump( RchartID, 0, 0, 0 );
#endif
                doUnknownWordProcessing( PchartID, RchartID, lastTailV );

#if defined(DEBUG) || defined(DEBUGU)
                MAchartDump( RchartID, 0, 0, 0 );
#endif
                MAchartGetIrrApplied( PchartID, &irrApplied );
                doPrintResult( RchartID, lastTailV, irrApplied, ch );
                MAchartDestroy( RchartID );
                RchartID = -1;
            }

            /* destroy S-Type chart, T-Type chart for new words
             */
            MAchartDestroy( SchartID );
            SchartID = -1;
            MAchartDestroy( TchartID );
            TchartID = -1;
#if defined(KTS)
            printf("\n");
            fflush(stdout);
#endif
        } /* of while(1) for one word */
        /* destroy S-Type chart */
        MAchartDestroy( SchartID );
        SchartID = -1;

#if defined(DEBUG)
        cacheDump();
#endif
    } /* of while(1) for sentence */

#if defined(LOG_STATISTIC)
  {
    FILE *fp;
    char s[100];

    fp = fopen("log/chart_usage", "a+");
    if( fp == NULL )
        MAerror( NOT, E_FILEOPEN, __FILE__, __LINE__, "log file for chart open is failed.\n" );
    else {
        timeStr( s );
        fprintf( fp, "%s - ", s);
        fprintf( fp, "%s (%s) - ", VERSION, REVDATE );
        fprintf( fp, "max %d chart is used\n", maxUsedChartNum );
        fclose(fp);
    }
  }
#endif

    MAKDestroy();
    exit(1);
}


static void prepareProcess( argc, argv )
    int argc;
    char *argv[];
{
    char* s;
    char cTableName[_POSIX_PATH_MAX];
    char mTableName[_POSIX_PATH_MAX];

    setFuncName( "prepareProcess" );

    // handling arguments
    if( argc == 1 ) { printUsage( argv[0] ); exit(1); }


    /* read environment variables */
    s = getenv( CACHE );
    if( s != NULL ) {
        if( s[0] = '0' ) isUsingCache = FALSE;
        else isUsingCache = TRUE;
    }

    s = getenv( DICT_DIR );
    if( s == NULL )
        MAerror( END, E_ENVVAR, __FILE__, __LINE__,
            "%s environment variable is not set.\n", DICT_DIR );
    strcpy( dictDir, s );

    s = getenv( TABLE_DIR );
    if( s == NULL )
        MAerror( END, E_ENVVAR, __FILE__, __LINE__,
            "%s environment variable is not set.\n", TABLE_DIR );
    strcpy( tableDir, s );

    s = getenv( CTABLE );
    if( s == NULL )
        MAerror( END, E_ENVVAR, __FILE__, __LINE__,
            "%s environment variable is not set.\n", CTABLE );
    sprintf( cTableName, "%s/%s", tableDir, s );

    s = getenv( MTABLE );
    if( s == NULL )
        MAerror( END, E_ENVVAR, __FILE__, __LINE__,
            "%s environment variable is not set.\n", MTABLE );
    sprintf( mTableName, "%s/%s", tableDir, s );


    // Initialization of all global data structures
    MAKInit();

    // NdictID, EdictID, UdictID is set by this function
    if( handleOptions( argc, argv ) != TRUE )
        exit(1);

    // check all dictionary is opened or not.
    checkDictOpen();

    // load other table
    loadConnectivityTable( cTableName );
    loadmTable( mTableName );

    debugMesg("CHECK : options are processed and preparing steps are done.\n");
}

static void checkDictOpen()
{
    if( NdictID == -1 )
        MAerror( END, E_NDICTOPEN, __FILE__, __LINE__,
          "Normal dictionary must be specified and opened.\n" );

    if( EdictID == -1 )
        MAerror( END, E_EDICTOPEN, __FILE__, __LINE__,
          "Exception dictionary must be specified and opened.\n" );

    if( UdictID == -1 )
        MAerror( END, E_UDICTOPEN, __FILE__, __LINE__,
          "Unknown word dictionary must be specified and opened.\n" );

    if( IdictID == -1 )
        MAerror( END, E_IDICTOPEN, __FILE__, __LINE__,
          "Irregular dictionary must be specified and opened.\n" );
}


/*--------------------------------------------------------------------
 *
 * Misc functions, handle options
 *
 *--------------------------------------------------------------------
 */

printUsage( fn )
    char* fn;
{
    printf("%s %s (%s)\n", fn, VERSION, REVDATE);
    printf("usage: %s [ cfneiuh ]\n", fn);
    printf("-c                  input from standard input\n");
    printf("-f data-file        input from data file contains set of sentences\n");
    printf("-n dict-name        normal dictionary name,       default : %s\n", DEFAULT_NDICTNAME );
    printf("-e dict-name        exception dictionary name,    default : %s\n", DEFAULT_EDICTNAME );
    printf("-i dict-name        irregular dictionary name,    default : %s\n", DEFAULT_IDICTNAME );
    printf("-u dict-name        unknown word dictionary name, default : %s\n", DEFAULT_UDICTNAME );
    printf("-h                  for this screen\n");
    printf("\n");
    printf("Environment variables must be set :\n");
    printf("%s - directory in which dictionary files are\n", DICT_DIR );
    printf("%s - directory in which table files are\n", TABLE_DIR );
    printf("%s - connectivity table in %s\n", CTABLE, TABLE_DIR );
    printf("%s - unknown word guessing table in %s\n", MTABLE, TABLE_DIR );
    printf("Environment variables can be set :\n");
    printf("%s - using cache? must be '0' or '1', default is '1' for using\n", CACHE );
    printf("\n");

    return FALSE;
}

char opt2onechar( s )
    char* s;
{
    static char *opts[] = { "-c", "-f", "-n", "-e", "-i", "-h", "-u", NULL };
    static char onechar[] = { 'c', 'f', 'n', 'e', 'i', 'h', 'u'};

    register int i;

    for( i=0; opts[i] != NULL; i++ )
        if( !strcmp( s, opts[i] ) ) return onechar[i];
}

handleOptions( argc, argv )
    int argc;
    char *argv[];
{
    register int i;
    char nDictName[_POSIX_PATH_MAX];
    char eDictName[_POSIX_PATH_MAX];
    char iDictName[_POSIX_PATH_MAX];
    char uDictName[_POSIX_PATH_MAX];

    sprintf( nDictName, "%s/%s", dictDir, DEFAULT_NDICTNAME );
    sprintf( eDictName, "%s/%s", dictDir, DEFAULT_EDICTNAME );
    sprintf( iDictName, "%s/%s", dictDir, DEFAULT_IDICTNAME );
    sprintf( uDictName, "%s/%s", dictDir, DEFAULT_UDICTNAME );

    for( i=1; i<argc; i++ ) {

        switch( opt2onechar( argv[i] ) ) {
        case 'n' :
            debugMesg("option : n\n");
            if( i+1 >= argc ) return printUsage( argv[0] );
            else {
                if( argv[++i][0] == '-' ) return printUsage( argv[0] );
                sprintf( nDictName, "%s/%s", dictDir, argv[i] );
            }
            break;

        case 'e' :
            debugMesg("option : e\n");
            if( i+1 >= argc ) return printUsage( argv[0] );
            else {
                if( argv[++i][0] == '-' ) return printUsage( argv[0] );
                sprintf( eDictName, "%s/%s", dictDir, argv[i] );
            }
            break;

        case 'u' :
            debugMesg("option : u\n");
            if( i+1 >= argc ) return printUsage( argv[0] );
            else {
                if( argv[++i][0] == '-' ) return printUsage( argv[0] );
                sprintf( uDictName, "%s/%s", dictDir, argv[i] );
            }
            break;

        case 'i' :
            debugMesg("option : i\n");
            if( i+1 >= argc ) return printUsage( argv[0] );
            else {
                if( argv[++i][0] == '-' ) return printUsage( argv[0] );
                sprintf( iDictName, "%s/%s", dictDir, argv[i] );
            }
            break;

        case 'c' :
            debugMesg("option : c\n");
            break;

        case 'f' :
            debugMesg("option : f\n");
            if( i+1 >= argc ) return printUsage( argv[0] );
            else {
                fpInput = fopen( argv[++i], "r" );
                if( fpInput == NULL )
                    MAerror( END, E_FILEOPEN, __FILE__, __LINE__,
                      "file open for input data is failed.\n" );

                tokenSetInput( fpInput );
                isFileOpened = TRUE;
            }
            break;

        case 'h' :
        default :
            return printUsage( argv[0] );
        }
    }

    NdictID = MAdictOpen( nDictName, O_RDONLY, 0644, dictTypeNormal );
    EdictID = MAdictOpen( eDictName, O_RDONLY, 0644, dictTypeExce );
    UdictID = MAdictOpen( uDictName, O_RDONLY, 0644, dictTypeUnknown );
    IdictID = MAdictOpen( iDictName, O_RDONLY, 0644, dictTypeIrregular );

    return TRUE;
}


/* load table ------------------------------------------------------*/

loadConnectivityTable( fname )
char *fname; 
{
    FILE *fp; 
    char in[100]; 
    int i; 

    fp = fopen(fname,"r");
    if( fp == NULL )
        MAerror( END, E_OPENCTABLE, __FILE__, __LINE__, "Connectivity table file %s open is failed.\n" );
 
    for (i = 0; i < NUM_POS; i++) {
        if( fgets(in, 100, fp) == NULL )
            MAerror( END, E_WRONGCTABLE, __FILE__, __LINE__, "Connectivity table'format is wrong.\n" );
        memcpy(cTable[i], in, NUM_POS);
    }

#if 0
   { register int i, j; 
     printf("Connectivity table --- BEGIN\n");
     for(i = 0 ; i < NUM_POS ; ++i) {
         for (j = 0 ; j < NUM_POS ; ++j)
             putchar(cTable[i][j]) ; 
         putchar('\n') ; 
     }
     printf("Connectivity table --- END\n");
   }
#endif
}

loadmTable( fname )
char *fname; 
{
    FILE *fp; 
    char in[100]; 
    int i; 

    fp = fopen(fname,"r");
    if( fp == NULL )
        MAerror( END, E_OPENMTABLE, __FILE__, __LINE__, "mTable file %s open is failed.\n" );
 
    for (i = 0; i < NUM_POS; i++) {
        if( fgets(in, 100, fp) == NULL )
            MAerror( END, E_WRONGCTABLE, __FILE__, __LINE__, "mTable'format is wrong.\n" );
        memcpy(mTable[i], in, NUM_POS);
    }

#if 0
   { register int i, j; 
     printf("mTable --- BEGIN\n");
     for(i = 0 ; i < NUM_POS ; ++i) {
         for (j = 0 ; j < NUM_POS ; ++j)
             putchar(mTable[i][j]) ; 
         putchar('\n') ; 
     }
     printf("mTable --- END\n");
   }
#endif
}

/* initialize and destroy ------------------------------------------*/

MAKInit()
{
    MAdictInit();    isInitDict = TRUE;
    tokenInit();     isInitToken = TRUE;
    MAchartInit();   isInitChart = TRUE;
    isInit();        isInitStack = TRUE;
    if( isUsingCache == TRUE ) cacheInit();
}

/* it must be provided some clearing routine to exit
 * gracefully, after clear out all allocated memory.
 * so in this routine, only call them.
 */
MAKDestroy()
{
    if( isInitToken == TRUE ) tokenDestroy();

    // close dictionary
    if( NdictID != -1 ) MAdictClose(NdictID);
    if( EdictID != -1 ) MAdictClose(EdictID);
    if( IdictID != -1 ) MAdictClose(IdictID);
    if( UdictID != -1 ) MAdictClose(UdictID);

    if( isFileOpened == TRUE ) fclose( fpInput );
    if( isUsingCache == TRUE ) cacheDestroy();
}
