/* dictionary API
 *
 * create : 95.5.26 (Fri), bgjang@csone
 * modify :
 * 95.5.27 (Sat) - make dictAPI.h and insert dictErrorStr
 * 95.5.28 (Sun) - fix some obvious bugs
 * 95.5.29 (Mon) - insert functions to support linear access to dict.
 *
 *
 * so generic and if dbm system is changed to DBMS, then
 * only the internal codes are changed.
 *
 * when error is occured :
 * 1. dictErrorStr is set by the contents by the error.
 * 2. error code is returned by the functions.
 */

#include <stdio.h>
#include <ndbm.h>
#include <dictAPI.h>


/* global variables definations ----------------------------------------*/

char* dictResult;          // this is an string, null-terminated
int dictResultLen;         // length of dictResult's content

char dictErrorStr[256];    // set when error is occured


/* static variables -----------------------------------------------------*/

static int dictResultSize;         // size of currently allocated dictResult
static DBM *id2dbm[MAX_OPEN_DICT]; // for, dict ID -> dbm pointer
                                   // ID : 0 .. MAX_OPEN_DICT-1


/* global variable basic interface --------------------------------------*/

dictResultSet( ptr, size )
    char* ptr;
    int size;
{
    // if size is not fit for current global var, then realloc
    if( dictResultSize < size ) {
        dictResult = (char*) realloc( dictResult, size );
        if( dictResult == NULL ) {
            sprintf(dictErrorStr, "%s : realloc() failed.\n", __FILE__);
            return D_ERR_REALLOC;
        }
        dictResultSize = size;
    }

    memcpy( dictResult, ptr, size );
    dictResultLen = size;
    return D_ERR_NO;
}


/* dictionary ADT -------------------------------------------------------*/

/* initialize dictionary API
 */
dictInit()
{
    /* init id2dbm array */
    register int i;
    for( i = 0; i < MAX_OPEN_DICT; i++ )
        id2dbm[i] = NULL;

    /* init global var */
    dictResultSize = 0;
    dictResult = (char*) malloc(INIT_RESULT_SIZE);
    if( dictResult == NULL ) {
        sprintf(dictErrorStr, "%s : malloc() failed.\n", __FILE__);
        return D_ERR_MALLOC;
    }
    dictResultSize = INIT_RESULT_SIZE;
    return D_ERR_NO;
}

/* open one dictionary
 */
dictOpen( path, flag, mode )
    char* path;
    int flag;
    int mode;
{
    register int i;
    DBM* db;

    /* find empty slot from id2dbm array */
    for( i = 0; i < MAX_OPEN_DICT; i++ )
	if( id2dbm[i] == NULL ) break;
    if( i == MAX_OPEN_DICT ) {
        sprintf(dictErrorStr, "%s : No more file can be opend.\n", __FILE__);
        return D_ERR_MAX_OPEN;
    }

    /* open dbm file */
    db = dbm_open( path, flag, mode );
    if( db == NULL ) {
        sprintf(dictErrorStr, "%s : file %s open failed.\n", __FILE__, path );
        return D_ERR_OPEN;
    }

    /* set id2dbm array and return dictionary ID */
    id2dbm[i] = db;
    return i;
}


/* close one dictionary
 */
dictClose( id )
    int id;
{
    DBM* db;
    int r;

    /* check ID and set DBM var */
    if( (r = checkID( id )) < 0 ) return r;
    db = id2dbm[id];

    dbm_close( id2dbm[id] );
    id2dbm[id] = NULL;
    return D_ERR_NO;
}

/* find by a keyword, and the result will be stored in
 * global variable.
 *
 * two normal return :
 *    D_NO_DATA - data is not existed.
 *    D_ERR_NO  - data is existed.
 */
dictFind( id, keyword )
    int id;
    char* keyword;
{
    DBM* db;
    datum key, data;
    int r;

    /* check ID and set DBM var */
    if( (r = checkID( id )) < 0 ) return r;
    db = id2dbm[id];

    /* fetch */
    key.dsize = strlen(keyword) + 1;
    key.dptr = keyword;
    data = dbm_fetch(db, key);

    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    /* if data is not existed */
    if( data.dptr == NULL ) return D_NO_DATA;

    /* global var set */
    r = dictResultSet( data.dptr, data.dsize );
    return r;
}

dictGetFirstWord( id, wordPtr )
    int id;
    char *wordPtr;  // OUT
{
    DBM* db;
    datum key, data;
    int r;

    /* check ID and set DBM var */
    if( (r = checkID( id )) < 0 ) return r;
    db = id2dbm[id];

    key = dbm_firstkey( db );
    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    if( key.dptr == NULL ) return D_NO_DATA;

    data = dbm_fetch(db, key);
    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    // if data is not existed
    if( data.dptr == NULL ) return D_ERR_NO_CONTENTS;

    strcpy( wordPtr, key.dptr );

    // global var set
    r = dictResultSet( data.dptr, data.dsize );
    return r;
}

dictGetNextWord( id, wordPtr )
    int id;
    char *wordPtr;   // OUT
{
    DBM* db;
    datum key, data;
    int r;

    /* check ID and set DBM var */
    if( (r = checkID( id )) < 0 ) return r;
    db = id2dbm[id];

    key = dbm_nextkey( db );
    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    if( key.dptr == NULL ) return D_NO_DATA;

    data = dbm_fetch(db, key);
    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    // if data is not existed
    if( data.dptr == NULL ) return D_ERR_NO_CONTENTS;

    strcpy( wordPtr, key.dptr );

    // global var set
    r = dictResultSet( data.dptr, data.dsize );
    return r;
}

/* insert one in global var to dictionary
 * two normal return :
 *     D_EXIST_DATA - data is existed, so return without modifications.
 */
dictInsert( id, keyword )
    int id;
    char* keyword;
{
    DBM* db;
    datum key, data;
    int r;

    /* check ID and set DBM var */
    if( (r = checkID( id )) < 0 ) return r;
    db = id2dbm[id];

    /* check whether exist or not */
    key.dsize = strlen(keyword) + 1;
    key.dptr = keyword;
    data = dbm_fetch(db, key);

    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    /* data existed, so return error */
    if( data.dptr != NULL ) {
        sprintf(dictErrorStr, "%s : data is existed in database.\n", __FILE__);
        return D_ERR_EXIST_DATA;
    }

    /* insert */
    data.dptr = dictResult;
    data.dsize = dictResultLen;
    r = dbm_store(db, key, data, DBM_INSERT);

    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    if( r < 0 ) {
        sprintf(dictErrorStr, "%s : insert failed.\n", __FILE__);
        return D_ERR_INSERT;
    }

    return D_ERR_NO;
}

dictModify( id, keyword )
    int id;
    char* keyword;
{
    DBM* db;
    datum key, data;
    int r;

    /* check ID and set DBM var */
    if( (r = checkID( id )) < 0 ) return r;
    db = id2dbm[id];

    /* check whether exist or not */
    key.dsize = strlen(keyword) + 1;
    key.dptr = keyword;
    data = dbm_fetch(db, key);

    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    /* data not existed, so return err */
    if( data.dptr == NULL ) {
        sprintf(dictErrorStr, "%s : data is not existed.\n", __FILE__);
        return D_ERR_NOT_EXIST;
    }

    /* replace */
    data.dptr = dictResult;
    data.dsize = dictResultLen;
    r = dbm_store(db, key, data, DBM_REPLACE);

    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    if( r < 0 ) {
        sprintf(dictErrorStr, "%s : modify failed.\n", __FILE__ );
        return D_ERR_MODIFY;
    }

    return D_ERR_NO;

}

/* delete one in dictionary
 */
dictDelete( id, keyword )
    int id;
    char* keyword;
{
    DBM* db;
    datum key, data;
    int r;

    /* check ID and set DBM var */
    if( (r = checkID( id )) < 0 ) return r;
    db = id2dbm[id];

    /* delete */
    key.dptr = keyword;
    key.dsize = strlen(keyword) + 1;
    r = dbm_delete(db, key);

    if( dbm_error(db) ) return dbmErrorResetAndReturn(db);

    if( r < 0 ) {
        sprintf(dictErrorStr, "%s : delete failed.\n", __FILE__);
        return D_ERR_DELETE;
    }

    return D_ERR_NO;
}

/* misc -------------------------------------------------------------*/

checkID( id )
    int id;
{
    if( id2dbm[id] == NULL ) {
        sprintf(dictErrorStr, "%s : Wrong id %d is passed to this function.\n", __FILE__, id);
        return D_ERR_WRONG_ID;
    }
    else return D_ERR_NO;
}

dbmErrorResetAndReturn(db)
    DBM* db;
{
    dbm_clearerr(db);
    sprintf(dictErrorStr, "%s : error in reading or writing database.\n", __FILE__);
    return D_ERR_R_OR_W;
}

/* EOF --------------------------------------------------------------*/
